
(module grammar-module mzscheme
  (provide sentence sentence2
           noun-phrase noun-phrase2 Noun 
           verb-phrase Verb
           Adj* PP* PP Adj Prep Article 
           word-list->string generate-and-speak 
          ) 
  
 ; (require (lib "plt-pretty-big.ss" "lang"))
  (require htdp/draw
           scheme/list)
  (require se3-bib/tools-module
           se3-bib/macos-module)
  
;;;; ===========================================================================
;;;; Generate sentences of a simple context free language, 
;;;;  see Norvig 92 for a Common Lisp Version
;;;; ===========================================================================

;;; A Grammar for a Subset of English
;;; adapted from Norvig-92, pp.35-42
;;;
;;; The Grammar
;;;  Sentence -> Noun-Phrase + Verb-Phrase
;;;  Noun-Phrase -> Article + Noun
;;;  Verb-Phrase -> Verb + Noun-Phrase
;;;  Article -> the, a, ...
;;;  Noun -> man, ball, woman, table, ...
;;;  Verb -> hit, took, saw, liked ...
;;;  

(define (sentence)
       (append (noun-phrase) (verb-phrase)))
(define (noun-phrase) 
       (append (Article) (Noun)))
(define (verb-phrase)
       (append (Verb) (noun-phrase)))
(define (Article)
       (one-of '(the a)))
(define (Noun)
       (one-of '(man ball woman table)))
(define (Verb) 
       (one-of '(hit took saw liked found)))

;;;; 2. Form
  (define *max-length* 30)
  
(define (sentence2)
       (append (noun-phrase2) (verb-phrase)))

(define (Adj*)
   (if (= (random 2) 0)
       '()
       (append (Adj) (Adj*))))
(define (PP*)
   (if (random-elt '(#t #f))
       (append (PP) (PP*))
       '()))

(define (noun-phrase2) 
   (append (Article) (Adj*) (Noun) (PP*)))

(define (PP)  (append (Prep) (noun-phrase2)))

(define (Adj) 
   (one-of '(big little blue green small)))
  
(define (Prep) (one-of '(to in by with on under))) 
  
(define (word-list->string words)
  (apply string-append 
         (map (lambda (word)
         (string-append 
          (symbol->string word) " "))
       words)))
  
(define (generate-and-speak phrase)
  (let ((thePhrase (phrase)))
    (display thePhrase)
  (Victoria (word-list->string 
             (take thePhrase 
                   (min *max-length* (length thePhrase )))
             ))))
  
;testen:
  ;(require trace)
  ;;(trace sentence noun-phrase verb-phrase article noun verb)
;;(untrace sentence noun-phrase verb-phrase article noun verb) 
  
  ; (generate-and-speak sentence)

  )