(module simActorClass-module swindle
  #|
 =================================================================
  sim-actor A mixin class for simulation objects,
  provides a framework for
      - broadcasting messages to all actors
      -initializing and resetting all actors
      - displaying a picture of the actors
 =================================================================
|#
  
  (provide 
   ; accessors
   actor-num actor-name sim-actor?
   theActors set-theActors! theLastNum
   ; global variables
   *theNumberGenerator* *all-actors*
   )
  
  
  #| auto provide 
  classes: sim-actor 
  generics: initialize
  sim-init! sim-start sim-info  
   tell broadcast broadcast-some
  add-actor!
|#
  
  (print "loading: simActorClass-module ")
  (require se3-bib/sim/simBase/sim-utility-base-package)
  
  (defclass* actorNumbers ()
    (lastnum :initvalue 0
             :accessor theLastNum
             :type <integer>
             :allocation :class)
    :autopred #t ; auto generate predicate sim-actor?
    :printer  #t 
    :documentation "a generator for unique actor numbers"
    )
  
  (define *theNumberGenerator* 
    (make actorNumbers))
  
  (defgeneric* next-actor-number ((ng actorNumbers))
    :documentation "generate a unique actor number"
    :method (method 
             ((ng actorNumbers))
             (inc! (theLastNum ng))
             (theLastNum ng)))
  
  (defmethod next-actor-number 
    ((ng actorNumbers))
    (inc! (theLastNum ng))
    (theLastNum ng))
  
  (defclass* sim-actor ()
    (actorName 
     :reader actor-name
     :initvalue "anonymous"
     :initarg :actorName
     :type <string>
     :documentation "The name of the actor"
     )
    (actorNum
     :reader actor-num
     :initializer 
     (lambda () 
       (next-actor-number 
        *theNumberGenerator*))
     :type <integer>
     :documentation 
     "A picture of the current state of the actor"
     )
    :autopred #t ; auto generate predicate sim-actor?
    :printer  #t 
    
    ;:documentation "A mixin class for simulation objects,
    ;provides a framework for broadcasting messages to all actors"
    )
  
  (defclass* setOfActors ()
    (theActors
     :accessor theActors
     :writer set-theActors!
     :initvalue '()
     :type <list>
     :documentation 
     "A set of actors"
     )
    :autopred #t ; auto generate predicate set-of-actors?
    :printer  #t 
    )
  
  (define *all-actors* (make setOfActors))
  
  (defgeneric* add-actor! ((a sim-actor) (s setOfActors))
    :documentation 
    "add an actor to the set of all actors"
    )
  
  (defgeneric* remove-duplicates! ((s setOfActors))
    :documentation 
    "remove duplicate entries"
    )
  
  (defgeneric* sim-init! ((obj sim-actor))
    :documentation 
    "further initializations after all actors have been created"
    :combination generic-begin-combination)
  
  (defgeneric* sim-start ((obj sim-actor))
    :documentation 
    "Trigger the start-up actions, get the system going"
    :combination generic-begin-combination)
  
  (defgeneric* sim-info ((obj sim-actor))
    :documentation 
    "request state information."
    :combination generic-begin-combination)
  
  
  ;;; ===========================================================
  ;;; Message passing
  ;;; ===========================================================
  
  (defgeneric* tell ((obj sim-actor) (message <function>))
    :documentation 
    "Pass message to actor. 
   The message must be the name 
   of an applicable method or procedure."
    )
  
  (defgeneric* broadcast 
    ((actor sim-actor) 
     (message <function>) 
     &key (sim-class sim-actor))
    :documentation 
    "Broadcast a message to all actors of type sim-type"
    )
  
  (defgeneric* broadcast-some 
    ((actor sim-actor) 
     (message <function>) 
     &optional (sim-type sim-actor))
    :documentation 
    "Send message to actors of type sim-type,
    and find some actor who confirms the message
    by answering 'not #f'."
    )
  
  
  )