(module the-towers-of-hanoi-module mzscheme
  (provide hanoi hanoiv stop *max-n-scheiben*) 
  
  (require lang/plt-pretty-big
           htdp/draw           
           se3-bib/tools-module
           se3-bib/my-vector-graphics)
  
  ;;; ===================================================================
  ;;; Die T?rme von Hanoi
  ;;;   mit Textausgabe
  ;;; ===================================================================
  
  (define (hanoi-r n start hilf ziel)
    (if (= n 1)
        (list (zug start ziel))
        (append 
         (hanoi-r (- n 1) start ziel hilf)
         (list (zug start ziel))
         (hanoi-r (- n 1) hilf start ziel))))
  
  (define (hanoi n) (hanoi-r n 1 2 3))
  
  (define (zug start ziel) (cons start ziel))
  
  ;;; ===================================================================
  ;;; Die T?rme von Hanoi
  ;;; mit Grafikausgabe
  ;;; ===================================================================
  
  (define *breite* 800) ; die Breite der Zeichenfl?che (pixel)
  (define *hoehe* 400)  ; die H?he der Zeichenfl?che (pixel)
  (define *rand* (* *hoehe* 0.1))
  (define *max-n-scheiben* 11)
  (define *max-scheibenbreite* (round (/ *breite* 3.1)))
  (define *scheibenhoehe*
    (round (/ (- *hoehe* *rand* *rand*) *max-n-scheiben*)))
  (define *hintergrundfarbe* 'white)
  (define *stabfarbe* 'black)
  (define *zahl-der-Scheiben* 1)
  
  
  
  (define (hanoi-g n start hilf ziel)
    ;returns the number of moves
    (if (= n 1)
        (zug-g! start ziel)
        (+ 
         (hanoi-g (- n 1) start ziel hilf)
         (zug-g! start ziel)
         (hanoi-g (- n 1) hilf start ziel))))
  
  (define (zug-g! start ziel)
    (let ((scheibe (wegnehmen! start)))
      (drauflegen! scheibe ziel)
      (if (< *zahl-der-Scheiben* 5) (sleep-for-a-while 1))
      1))
  
  ;; abstrakter Datentyp f?r T?rme:
  ;; realisiert als structur: erstes Element = Turmnummer, 
  ;;                       zweites Element = Liste der Scheiben
  ;; nicht referentiell transparent: die Funktionen drauflegen!, wegnehmen!
  ;;                       ver?ndern den Zustand der T?rme!
  (define-struct turm 
  (nummer scheiben))
  
  (define (neuer-turm turmnummer) 
    (make-turm
      turmnummer '()))
  
  (define (AlleScheibenAuflegen! turm n)
    (let ((dieScheiben (if (> n 0)
                           (reverse (nats-1-n n)) '())))
      (map (rcurry drauflegen! turm) dieScheiben)
      turm))
  
  (define (turmnummer turm) 
    (turm-nummer turm))
  
  (define (scheiben turm) 
    (turm-scheiben turm))
  
  (define (hoehe turm) (length (scheiben turm)))
  
  (define (turmspitze-x turm)
    ; bezogen auf das Bildschirmystem, Ursprung links oben
    (round (* (/ *breite* 3)
              (- (turmnummer turm) 0.5) )))
  
  (define (turmspitze-y turm)
    ; bezogen auf das  Bildschirmsystem, Ursprung links oben
    (round (- *hoehe* *rand* (* (hoehe turm) *scheibenhoehe*))))
  
  
  (define (farbe scheibe)
    (let ((farben '(green red blue yellow)))
      (list-ref farben (remainder scheibe (length farben)))))
  
  (define (breite scheibe)
    (round (* scheibe *max-scheibenbreite* (/ 1 *max-n-scheiben*))))
  
  (define (drauflegen! scheibe turm)
    (let ((diescheiben (cons scheibe (scheiben turm))))
      ;(display (list 'drauflegen scheibe 'turm (turmnummer turm)))
      (set-turm-scheiben! turm diescheiben)
      (zeichne scheibe turm)
      turm
      ))
  
  (define (wegnehmen! turm)
    (let ((scheibe (car (scheiben turm))))
      ;(display (list 'wegnehmen scheibe 'turm (turmnummer turm)))
      (loesche scheibe turm) 
      (set-turm-scheiben! turm (cdr (turm-scheiben turm)))
      scheibe))
  
  
  (define (zeichne scheibe turm)
    (let* ((scheibenbreite (breite scheibe))
           (x (- (turmspitze-x turm) (/ scheibenbreite 2)))
           (y (turmspitze-y turm))
           (scheibenfarbe (farbe scheibe))
           (pos (make-posn (round x) (round y))))
      (draw-solid-rect pos 
                       scheibenbreite *scheibenhoehe* scheibenfarbe)
      ))
  
  (define (loesche scheibe turm)
    (let* ((scheibenbreite (breite scheibe))
           (xstab (turmspitze-x turm))
           (xscheibe (- xstab (/ scheibenbreite 2)))
           (yscheibe (turmspitze-y turm))
           (posscheibe (make-posn (round xscheibe) (round yscheibe)))
           (posstabunten (make-posn (round xstab) (round yscheibe)))
           (posstaboben 
            (make-posn 
             (round xstab) 
             (round (+ yscheibe *scheibenhoehe*)))))
      (draw-solid-rect posscheibe 
                       scheibenbreite *scheibenhoehe* *hintergrundfarbe*)
      (draw-solid-line  posstaboben posstabunten *stabfarbe*)
      
      ))
  
  (define (zeichneStab turm n)
    (let* ((xstab (turmspitze-x turm))
           (posstabunten (make-posn (round xstab) (- *hoehe* *rand*)))
           (posstaboben 
            (make-posn 
             (round xstab) 
             (round (- *hoehe* *rand* (* (+ 1 n) *scheibenhoehe*))))))
      (draw-solid-line  posstaboben posstabunten *stabfarbe*) ))
  
  (define (zeichneGestell n turm1 turm2 turm3)
    (let* ((x 0)
           (y (- *hoehe* *rand*))
           (pos (make-posn (round x) (round y))))
      (draw-solid-rect pos *breite* *rand* *stabfarbe*)
      (zeichneStab turm1 n)
      (zeichneStab turm2 n)
      (zeichneStab turm3 n)
      ))
  
  (define (hanoiv n)
    ; zeige die Z?ge graphisch
    (set! *zahl-der-Scheiben* n)
    (if (and (integer? n) (<= 1 n *max-n-scheiben*))
        (begin
          (let ((ausgangsturm (neuer-turm 1));  anfangs leer
                (hilfsturm (neuer-turm 2)); anfangs leer
                (zielturm (neuer-turm 3))); anfangs leer
            
            (start *breite* *hoehe*); create canvas
            (zeichneGestell n ausgangsturm hilfsturm zielturm)
            (AlleScheibenAuflegen! ausgangsturm n) ;alle Scheiben anfangs hier
            (hanoi-g n ausgangsturm hilfsturm zielturm))
          ); close canvas
        (begin (display (string-append 
                         "n mu? eine nat?rliche Zahl zwischen 1 und "
                         (number->string *max-n-scheiben*) " sein!"))
               #f   )))
  ; Test: (require (lib "the-towers-of-hanoi-module.ss" "teachpack" "se3-bib"))
  ;(hanoiv 7)
  )
  ; end module the-towers-of-hanoi 

